%{
/*
 * Copyright (c) 2004 - 2013 Marc Balmer <marc@msys.ch>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <err.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>

#include "smtp-vilter.h"
#include "pathnames.h"

#define MAXTRIES	256

extern FILE	*icapin;
extern int	 icaplineno;
extern char	*icaptext;

extern int	 icap_port;
extern char	*icap_host;
extern char	*bind_addr;
extern char	*icap_uri;
extern char	*icap_query;
extern int	 icap_tries;
extern int	 icap_timeout;

static int	 icaperrcnt;
static char	*icapcfgfile;

int	icaperror(const char *, ...);
int	icapparse(void);
int	icaplex(void);
%}

%union {
	long number;
	char *string;
}

%token	HOST PORT BIND URI QUERY TRIES TIMEOUT
%token	<string>	NUMBER
%token	<string>	TEXT

%%
statement	: /* empty */
		| statement '\n'
		| statement host '\n'
		| statement port '\n'
		| statement bind '\n'
		| statement uri '\n'
		| statement query '\n'
		| statement tries '\n'
		| statement timeout '\n'
		;

host		: HOST '=' TEXT				{
			free(icap_host);
			icap_host = $3;
		}
		;

port		: PORT '=' NUMBER			{
			icap_port = atoi($3);
			free($3);
		}
		;

bind		: BIND '=' TEXT				{
			if (bind_addr != NULL)
				free(bind_addr);
			bind_addr = $3;
		}
		;

uri		: URI '=' TEXT				{
			free(icap_uri);
			icap_uri = $3;
		}
		;

query		: QUERY '=' TEXT			{
			free(icap_query);
			icap_query = $3;
		}
		;

tries		: TRIES '=' NUMBER			{
			icap_tries = atoi($3);
			free($3);
		}
		;
timeout		: TIMEOUT '=' NUMBER			{
			icap_timeout = atoi($3);
			free($3);
		}
		;


%%

int
vilter_init(char *cfgfile)
{
	if (verbose)
		warnx("icap: vilter_init()");

	/* Default values */

	icap_port = 1344;
	bind_addr = NULL;

	icap_host = strdup("localhost");
	icap_uri = strdup("icap://localhost(");
	icap_query = strdup("action=SCAN");

	if (icap_host == NULL || icap_uri == NULL || icap_query == NULL)
		err(1, "icap: out of memory");

	icap_tries = 1;
	icap_timeout = 0;

	if (cfgfile == NULL)
		cfgfile = ICAP_CONF;

	icaplineno = 1;
	icaperrcnt = 0;
	icapcfgfile = cfgfile;
	if ((icapin = fopen(cfgfile, "r")) != NULL) {
		while (!feof(icapin))
			icapparse();
		fclose(icapin);
	} else if (verbose)
		warnx("icap: configuration file %s for icap backend not "
		    "found, using default values", cfgfile);

	if (icaperrcnt)
		errx(1, "configuration file contains errors, terminating");

	return 0;
}

int
icaperror(const char *fmt, ...)
{
	va_list		 ap;
	char		*nfmt;

	++icaperrcnt;
	va_start(ap, fmt);
	if (asprintf(&nfmt, "%s, line %d: %s near '%s'",
	    icapcfgfile, icaplineno, fmt, icaptext) == -1)
		errx(1, "icap: error asprintf");
	va_end(ap);
	fprintf(stderr, "%s\n", nfmt);
	free(nfmt);
	return 0;
}
